from __future__ import annotations

from typing import TYPE_CHECKING

from dagster import ConfigurableIOManager, InputContext, OutputContext

DAGSTER_RAY_OBJECT_MAP_NAME = "DagsterRayObjectMap"
DAGSTER_RAY_NAMESPACE = "dagster-ray"

# we need to create a global Ray actor which will store all the refs to all objcets

if TYPE_CHECKING:
    import ray


class RayObjectMap:
    # TODO: implement some eventual cleanup mechanism
    # idea: save creation timestamp and periodically check for old refs
    # or add some integration with the RunLauncher/Executor
    def __init__(self):
        self._object_map: dict[str, ray.ObjectRef] = {}

    def set(self, key: str, ref: ray.ObjectRef):
        self._object_map[key] = ref

    def get(self, key: str) -> ray.ObjectRef | None:
        return self._object_map.get(key)

    def delete(self, key: str):
        if key in self._object_map:
            del self._object_map[key]

    def keys(self):
        return self._object_map.keys()

    def ping(self):
        return "pong"

    @staticmethod
    def get_or_create():
        import ray

        actor = (
            ray.remote(RayObjectMap)
            .options(  # type: ignore
                name=DAGSTER_RAY_OBJECT_MAP_NAME,
                namespace=DAGSTER_RAY_NAMESPACE,
                get_if_exists=True,
                lifetime="detached",
                # max_restarts=-1,
                max_concurrency=1000,  # TODO: make this configurable,
                runtime_env={"RAY_ENABLE_RECORD_ACTOR_TASK_LOGGING": "1"},
            )
            .remote()
        )

        # make sure the actor is created
        ray.get(actor.ping.remote())  # type: ignore

        return actor


class RayIOManager(ConfigurableIOManager):
    address: str | None = None

    def handle_output(self, context: OutputContext, obj):
        import ray

        if self.address:  # TODO: should this really be done here?
            ray.init(self.address, ignore_reinit_error=True)

        object_map = RayObjectMap.get_or_create()

        storage_key = self._get_single_key(context)

        # TODO: understand if Ray will automatically move the object from dying nodes
        # what if not?

        ref = ray.put(obj, _owner=object_map)

        object_map.set.remote(storage_key, ref)  # type: ignore

        context.log.debug(f"[RayIOManager] Stored object with key {storage_key} as {ref}")

    def load_input(self, context: InputContext):
        import ray

        if self.address:  # TODO: should this really be done here?
            ray.init(self.address, ignore_reinit_error=True)

        object_map = RayObjectMap.get_or_create()

        if context.has_asset_partitions and len(context.asset_partition_keys) > 1:
            # load multiple partitions as once
            # first, get the refs

            storage_keys = self._get_multiple_keys(context)
            refs = [object_map.get.remote(key) for key in storage_keys.values()]  # type: ignore
            values = ray.get(refs)
            return {partition_key: value for partition_key, value in zip(storage_keys.keys(), values)}

        else:
            storage_key = self._get_single_key(context)

        context.log.debug(f"[RayIOManager] Loading object with key {storage_key}")

        ref = object_map.get.remote(storage_key)  # type: ignore

        assert ref is not None, f"[RayIOManager] Object with key {storage_key} not found in RayObjectMap"

        return ray.get(ref)

    def _get_single_key(self, context: InputContext | OutputContext) -> str:
        identifier = context.get_identifier() if not context.has_asset_key else context.get_asset_identifier()
        return "/".join(identifier)

    def _get_multiple_keys(self, context: InputContext) -> dict[str, str]:
        if context.has_asset_key:
            asset_path = list(context.asset_key.path)

            return {
                partition_key: "/".join(asset_path + [partition_key]) for partition_key in context.asset_partition_keys
            }
        else:
            raise RuntimeError(
                "[RayIOManager] This method can only be called with an InputContext that has multiple partitions"
            )



from __future__ import annotations

import time
from typing import TYPE_CHECKING, Any, Generic, TypeVar

if TYPE_CHECKING:
    from kubernetes import client
    from kubernetes.client.models.v1_endpoints import V1Endpoints


def load_kubeconfig(context: str | None = None, config_file: str | None = None) -> Any:
    from kubernetes import config

    try:
        return config.load_kube_config(context=context, config_file=config_file)
    except config.config_exception.ConfigException:
        try:
            return config.load_incluster_config()
        except config.config_exception.ConfigException:
            pass


T_Status = TypeVar("T_Status")


class BaseKubeRayClient(Generic[T_Status]):
    def __init__(
        self,
        group: str,
        version: str,
        kind: str,
        plural: str,
        api_client: client.ApiClient | None = None,
    ):
        from kubernetes import client

        self.group = group
        self.version = version
        self.kind = kind
        self.plural = plural
        self.api_client = api_client
        self._api = client.CustomObjectsApi(api_client=api_client)
        self._core_v1_api = client.CoreV1Api(api_client=api_client)

    def wait_for_service_endpoints(self, service_name: str, namespace: str, poll_interval: int = 5, timeout: int = 600):
        from kubernetes.client import ApiException

        start_time = time.time()

        while True:
            try:
                # Get the endpoints for the service
                endpoints: V1Endpoints = self._core_v1_api.read_namespaced_endpoints(service_name, namespace)  # type: ignore

                # Check if there are addresses in any of the subsets
                if endpoints.subsets:
                    for subset in endpoints.subsets:
                        if subset.addresses:
                            return
            except ApiException:
                pass

            elapsed_time = time.time() - start_time
            if elapsed_time > timeout:
                raise TimeoutError(
                    f"Timed out waiting for endpoints for service {service_name} in namespace {namespace}"
                )

            time.sleep(poll_interval)

    def get_status(self, name: str, namespace: str, timeout: int = 60, poll_interval: int = 5) -> T_Status:
        from kubernetes.client import ApiException

        while timeout > 0:
            try:
                resource: Any = self._api.get_namespaced_custom_object_status(
                    group=self.group,
                    version=self.version,
                    plural=self.plural,
                    name=name,
                    namespace=namespace,
                )
            except ApiException:
                raise

            if resource.get("status"):
                return resource["status"]
            else:
                time.sleep(poll_interval)
                timeout -= poll_interval

        raise TimeoutError(f"Timed out waiting for status of {self.kind} {name} in namespace {namespace}")

    def list(self, namespace: str, label_selector: str = "", async_req: bool = False) -> dict[str, Any]:
        from kubernetes.client import ApiException

        try:
            resource: Any = self._api.list_namespaced_custom_object(
                group=self.group,
                version=self.version,
                plural=self.plural,
                namespace=namespace,
                label_selector=label_selector,
                async_req=async_req,
            )
            if "items" in resource:
                return resource
            else:
                return {}
        except ApiException as e:
            if e.status == 404:
                return {}

            raise

    def get(self, name: str, namespace: str) -> dict[str, Any]:
        from kubernetes.client import ApiException

        try:
            resource: Any = self._api.get_namespaced_custom_object(
                group=self.group,
                version=self.version,
                plural=self.plural,
                name=name,
                namespace=namespace,
            )
            return resource
        except ApiException as e:
            if e.status == 404:
                return {}
            raise

    def create(self, body: Any, namespace: str) -> Any:
        return self._api.create_namespaced_custom_object(
            group=self.group,
            version=self.version,
            plural=self.plural,
            body=body,
            namespace=namespace,
        )

    def delete(self, name: str, namespace: str):
        return self._api.delete_namespaced_custom_object(
            group=self.group,
            version=self.version,
            plural=self.plural,
            name=name,
            namespace=namespace,
        )

    def uodate(self, name: str, ray_patch: Any, namespace: str):
        return self._api.patch_namespaced_custom_object(
            group=self.group,
            version=self.version,
            plural=self.plural,
            name=name,
            body=ray_patch,
            namespace=namespace,
        )



import re


def normalize_k8s_label_values(labels: dict[str, str]) -> dict[str, str]:
    # prevent errors like:
    # Invalid value: \"daniel@anam.ai\": a valid label must be an empty string or consist of alphanumeric characters, '-', '_' or '.', and must start and end with an alphanumeric character (e.g. 'MyValue',  or 'my_value',  or '12345', regex used for validation is '(([A-Za-z0-9][-A-Za-z0-9_.]*)?[A-Za-z0-9])?')
    # and comply with the 63 character limit

    cleanup_regex = re.compile(r"[^a-zA-Z0-9-_.]+")

    banned_starting_characters = ["-", "_", "."]

    for key, value in labels.items():
        # -daniel~!@my.domain -> daniel-my-domain
        with_maybe_bad_start = cleanup_regex.sub("", value.replace("@", "-").replace(".", "-"))
        while with_maybe_bad_start and with_maybe_bad_start[0] in banned_starting_characters:
            with_maybe_bad_start = with_maybe_bad_start[1:]
        labels[key] = with_maybe_bad_start[:63]

    return labels



from argparse import ArgumentParser

from packaging.version import Version

parser = ArgumentParser()

parser.add_argument("--python", type=Version, help="Python version")
parser.add_argument("--dagster", type=Version, help="Dagster version")
parser.add_argument("--ray", type=Version, help="Dagster version")


def check_python_312(
    python: Version,
    dagster: Version,
    ray: Version,
):
    if python >= Version("3.12") and ray < Version("2.37.0"):
        raise RuntimeError("Ray version must be >=2.37.0 for Python >=3.12")


def main():
    args = parser.parse_args()
    check_python_312(
        python=args.python,
        dagster=args.dagster,
        ray=args.ray,
    )


if __name__ == "__main__":
    main()



from __future__ import annotations

import os
from typing import cast

from dagster import AssetExecutionContext, InitResourceContext, OpExecutionContext

from dagster_ray._base.constants import DEFAULT_DEPLOYMENT_NAME


def get_dagster_tags(context: InitResourceContext | OpExecutionContext | AssetExecutionContext) -> dict[str, str]:
    """
    Returns a dictionary with common Dagster tags.
    """
    assert context.dagster_run is not None

    if isinstance(context, InitResourceContext):
        run_id = cast(str, context.run_id)
    else:
        run_id = context.run.run_id

    labels = {
        "dagster.io/run_id": run_id,
        "dagster.io/deployment": DEFAULT_DEPLOYMENT_NAME,
        # TODO: add more labels
    }

    if context.dagster_run.tags.get("user"):
        labels["dagster.io/user"] = context.dagster_run.tags["user"]

    if os.getenv("DAGSTER_CLOUD_GIT_BRANCH"):
        labels["dagster.io/git-branch"] = os.environ["DAGSTER_CLOUD_GIT_BRANCH"]

    if os.getenv("DAGSTER_CLOUD_GIT_SHA"):
        labels["dagster.io/git-sha"] = os.environ["DAGSTER_CLOUD_GIT_SHA"]

    return labels



from __future__ import annotations

import logging
import time
from collections.abc import Iterator
from typing import TYPE_CHECKING, Literal, TypedDict

from typing_extensions import NotRequired

from dagster_ray.kuberay.client.base import BaseKubeRayClient, load_kubeconfig
from dagster_ray.kuberay.client.raycluster import RayClusterClient, RayClusterStatus

if TYPE_CHECKING:
    from kubernetes.client import ApiClient

GROUP = "ray.io"
VERSION = "v1"
PLURAL = "rayjobs"
KIND = "RayJob"

logger = logging.getLogger(__name__)


class RayJobStatus(TypedDict):
    jobId: str
    jobDeploymentStatus: str
    rayClusterName: str
    rayClusterStatus: RayClusterStatus
    startTime: str

    dashboardURL: NotRequired[str]
    endTime: NotRequired[str]
    jobStatus: NotRequired[Literal["PENDING", "RUNNING", "SUCCEEDED", "FAILED", "STOPPED"]]
    message: NotRequired[str]


class RayJobClient(BaseKubeRayClient[RayJobStatus]):
    def __init__(
        self,
        config_file: str | None = None,
        context: str | None = None,
        api_client: ApiClient | None = None,
    ) -> None:
        # this call must happen BEFORE creating K8s apis
        load_kubeconfig(config_file=config_file, context=context)

        self.config_file = config_file
        self.context = context

        super().__init__(group=GROUP, version=VERSION, kind=KIND, plural=PLURAL, api_client=api_client)

    def get_ray_cluster_name(self, name: str, namespace: str) -> str:
        return self.get_status(name, namespace)["rayClusterName"]

    def get_job_sumission_id(self, name: str, namespace: str) -> str:
        return self.get_status(name, namespace)["jobId"]

    @property
    def ray_cluster_client(self) -> RayClusterClient:
        return RayClusterClient(config_file=self.config_file, context=self.context)

    def wait_until_running(
        self,
        name: str,
        namespace: str,
        timeout: int = 600,
        poll_interval: int = 5,
        terminate_on_timeout: bool = True,
        port_forward: bool = False,
    ) -> bool:
        start_time = time.time()

        while True:
            status = self.get_status(name, namespace, timeout, poll_interval).get("jobDeploymentStatus")

            if status in ["Running", "Complete"]:
                break
            elif status == "Failed":
                raise RuntimeError(f"RayJob {namespace}/{name} deployment failed. Status:\n{status}")

            if time.time() - start_time > timeout:
                if terminate_on_timeout:
                    logger.warning(f"Terminating RayJob {namespace}/{name} because of timeout {timeout}s")
                    try:
                        self.terminate(name, namespace, port_forward=port_forward)
                    except Exception as e:
                        logger.warning(
                            f"Failed to gracefully terminate RayJob {namespace}/{name}: {e}, will delete it instead."
                        )
                        self.delete(name, namespace)

                raise TimeoutError(f"Timed out waiting for RayJob {namespace}/{name} to start. Status:\n{status}")

            time.sleep(poll_interval)

        while True:
            status = self.get_status(name, namespace, timeout, poll_interval).get("jobStatus")

            if status:
                break

            if time.time() - start_time > timeout:
                raise TimeoutError(f"Timed out waiting for RayJob {namespace}/{name} to start. Status:\n{status}")

            time.sleep(poll_interval)

        return True

    def _wait_for_job_submission(
        self,
        name: str,
        namespace: str,
        timeout: int = 600,
        poll_interval: int = 10,
    ):
        start_time = time.time()

        while True:
            status = self.get_status(name, namespace)
            if status.get("jobDeploymentStatus") in ["Complete", "Failed"]:
                return

            if (job_status := status.get("jobStatus")) is not None:
                if job_status != "PENDING":
                    return

            if time.time() - start_time > timeout:
                raise TimeoutError(f"Timed out waiting for job {name} to start")

            logger.debug(f"RayJob {namespace}/{name} deployment status is {job_status}, waiting for it to start...")

            time.sleep(poll_interval)

    def get_job_logs(self, name: str, namespace: str, timeout: int = 60 * 60, port_forward: bool = False) -> str:
        self._wait_for_job_submission(name, namespace, timeout=timeout)
        with self.ray_cluster_client.job_submission_client(
            name=self.get_ray_cluster_name(name, namespace), namespace=namespace, port_forward=port_forward
        ) as job_submission_client:
            return job_submission_client.get_job_logs(job_id=self.get_job_sumission_id(name, namespace))

    def tail_job_logs(
        self, name: str, namespace: str, timeout: int = 60 * 60, port_forward: bool = False
    ) -> Iterator[str]:
        import asyncio

        self._wait_for_job_submission(name, namespace, timeout=timeout)
        with self.ray_cluster_client.job_submission_client(
            name=self.get_ray_cluster_name(name, namespace), namespace=namespace, port_forward=port_forward
        ) as job_submission_client:
            async_tailer = job_submission_client.tail_job_logs(job_id=self.get_job_sumission_id(name, namespace))

            # Backward compatible sync generator
            def tail_logs() -> Iterator[str]:
                while True:
                    try:
                        yield asyncio.get_event_loop().run_until_complete(async_tailer.__anext__())  # type: ignore
                    except StopAsyncIteration:
                        break

            yield from tail_logs()

    def terminate(self, name: str, namespace: str, port_forward: bool = False) -> bool:
        """
        Unlike the .delete method, this won't remove the Kubernetes object, but will instead stop the Ray Job.
        """
        with self.ray_cluster_client.job_submission_client(
            name=self.get_ray_cluster_name(name, namespace), namespace=namespace, port_forward=port_forward
        ) as job_submission_client:
            job_id = self.get_job_sumission_id(name, namespace)

            job_submitted = False

            while not job_submitted:
                jobs = job_submission_client.list_jobs()

                for job in jobs:
                    if job.submission_id == job_id:
                        job_submitted = True
                        break

                logger.debug(
                    f"Trying to terminate job {name}, but it wasn't submitted yet. Waiting for it to be submitted..."
                )
                time.sleep(10)

            return job_submission_client.stop_job(job_id=job_id)



import os
from typing import Any

import dagster as dg
import pytest

from dagster_ray import LocalRay, RayResource


def test_runtime_env():
    import ray

    ray.init

    @ray.remote
    def my_func():
        assert os.environ["FOO"] == "BAR"

    @dg.asset
    def my_asset(ray_cluster: RayResource) -> None:
        ray.get(my_func.remote())

    ray_cluster = LocalRay(ray_init_options={"runtime_env": {"env_vars": {"FOO": "BAR"}}})

    with dg.DagsterInstance.ephemeral() as instance:
        dg.materialize(
            assets=[my_asset],
            instance=instance,
            resources={"ray_cluster": ray_cluster},
        )


def test_debug_mode():
    import ray

    ray.init

    @ray.remote
    def my_func():
        assert os.environ["RAY_PROFILING"] == "1"
        assert os.environ["RAY_ENABLE_RECORD_ACTOR_TASK_LOGGING"] == "1"
        assert os.environ["RAY_DEBUG_POST_MORTEM"] == "1"

    @dg.asset
    def my_asset(ray_cluster: RayResource) -> None:
        ray.get(my_func.remote())

    ray_cluster = LocalRay(enable_tracing=True, enable_actor_task_logging=True, enable_debug_post_mortem=True)

    with dg.DagsterInstance.ephemeral() as instance:
        dg.materialize(
            assets=[my_asset],
            instance=instance,
            resources={"ray_cluster": ray_cluster},
        )


@pytest.mark.parametrize(
    "ray_init_options",
    [
        {"runtime_env": {"env_vars": {"FOO": dg.EnvVar("BAR")}}},
        {"runtime_env": {"env_vars": {"FOO": {"env": "BAR"}}}},
        {"runtime_env": {"env_vars": {"FOO": dg.EnvVar("BAR"), "BAZ": dg.EnvVar("QUIX")}}},
        {"runtime_env": {"env_vars": {"FOO": dg.EnvVar("BAR"), "BAZ": {"env": "QUIX"}}}},
    ],
)
def test_runtime_env_env_var(ray_init_options: dict[str, Any]):
    import ray

    ray.init

    @ray.remote
    def my_func():
        assert os.environ["FOO"] == "BAR"

    @dg.asset
    def my_asset(ray_cluster: RayResource) -> None:
        ray.get(my_func.remote())

    os.environ["BAR"] = "BAR"

    with dg.DagsterInstance.ephemeral() as instance:
        dg.materialize(
            assets=[my_asset],
            instance=instance,
            resources={"ray_cluster": LocalRay(ray_init_options=ray_init_options)},
        )



from collections.abc import Iterator

import pytest
from dagster import (
    DagsterInstance,
    execute_job,
    instance_for_test,
    job,
    op,
    reconstructable,
)

from dagster_ray import RayIOManager
from dagster_ray.executor import ray_executor


@op
def return_one() -> int:
    return 1


@op
def return_two() -> int:
    return 2


@op
def sum_one_and_two(a: int, b: int) -> int:
    res = a + b

    assert res == 3

    return res


@job(executor_def=ray_executor)
def my_job():
    return_two_result = return_two()
    return_one_result = return_one()
    sum_one_and_two(return_one_result, return_two_result)


@op
def failed_op() -> None:
    raise RuntimeError("This op failed!")


@job(executor_def=ray_executor)
def my_failing_job():
    failed_op()


@pytest.fixture
def dagster_instance() -> Iterator[DagsterInstance]:
    with instance_for_test() as instance:
        yield instance


def test_ray_executor(local_ray_address: str, dagster_instance: DagsterInstance):
    result = execute_job(
        job=reconstructable(my_job),
        instance=dagster_instance,
        run_config={
            "execution": {
                "config": {
                    "ray": {"address": local_ray_address},
                }
            }
        },
    )

    assert result.success, result.get_step_failure_events()[0].event_specific_data


ray_io_manager = RayIOManager()


@job(executor_def=ray_executor, resource_defs={"io_manager": ray_io_manager})
def my_job_with_ray_io_manager():
    return_two_result = return_two()
    return_one_result = return_one()
    sum_one_and_two(return_one_result, return_two_result)


def test_ray_executor_with_ray_io_manager(local_ray_address: str, dagster_instance: DagsterInstance):
    result = execute_job(
        job=reconstructable(my_job_with_ray_io_manager),
        instance=dagster_instance,
        run_config={
            "execution": {
                "config": {
                    "ray": {"address": local_ray_address},
                }
            }
        },
    )

    assert result.success, result.get_step_failure_events()[0].event_specific_data


def test_ray_executor_local_failing(local_ray_address: str, dagster_instance: DagsterInstance):
    result = execute_job(
        job=reconstructable(my_failing_job),
        instance=dagster_instance,
        run_config={
            "execution": {
                "config": {
                    "ray": {"address": local_ray_address},
                }
            }
        },
    )
    assert result.get_failed_step_keys() == {"failed_op"}


RUNTIME_ENV = {"pip": ["polars"], "env_vars": {"FOO": "bar"}}


def runtime_env_checks():
    import polars  # noqa  # type: ignore
    import os

    assert os.environ["FOO"] == "bar"


@op
def op_testing_runtime_env():
    runtime_env_checks()


@job(executor_def=ray_executor)
def job_testing_runtime_env():
    op_testing_runtime_env()


def test_ray_executor_local_runtime_env(local_ray_address: str, dagster_instance: DagsterInstance):
    # first test this runtime_env just with ray

    import ray

    ray.get(ray.remote(runtime_env=RUNTIME_ENV)(runtime_env_checks).remote())

    result = execute_job(
        job=reconstructable(job_testing_runtime_env),
        instance=dagster_instance,
        run_config={
            "execution": {
                "config": {
                    "ray": {"address": local_ray_address, "runtime_env": RUNTIME_ENV},
                }
            }
        },
    )
    assert result.success, result.get_step_failure_events()[0].event_specific_data


@op(tags={"dagster-ray/config": {"runtime_env": RUNTIME_ENV}})
def op_with_user_provided_runtime_env():
    runtime_env_checks()


@job(executor_def=ray_executor)
def job_with_user_provided_runtime_env():
    op_with_user_provided_runtime_env()


def test_ray_executor_local_user_provided_runtime_env(local_ray_address: str, dagster_instance: DagsterInstance):
    # first test this runtime_env just with ray

    import ray

    ray.get(ray.remote(runtime_env=RUNTIME_ENV)(runtime_env_checks).remote())

    result = execute_job(
        job=reconstructable(job_with_user_provided_runtime_env),
        instance=dagster_instance,
        run_config={
            "execution": {
                "config": {
                    "ray": {"address": local_ray_address},
                }
            }
        },
    )
    assert result.success, result.get_step_failure_events()[0].event_specific_data
